# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2.

import asyncio
import itertools
import json
from dataclasses import dataclass
from typing import Dict, Generic, List, Optional, TypeVar, Union

from edenscm import error
from edenscm.i18n import _


T = TypeVar("T")


@dataclass
class Result(Generic[T]):
    ok: Optional[T] = None
    error: Optional[str] = None

    def is_error(self) -> bool:
        return self.error is not None


async def make_request(
    params: Dict[str, Union[str, int, bool]],
    hostname: str,
    endpoint="graphql",
    method: Optional[str] = None,
) -> Result:
    """If successful, returns a Result whose value is parsed JSON returned by
    the request.
    """
    if method:
        endpoint_args = ["-X", method.upper(), endpoint]
    else:
        endpoint_args = [endpoint]
    args = (
        ["gh", "api", "--hostname", hostname]
        + endpoint_args
        + list(itertools.chain(*[_format_param(k, v) for (k, v) in params.items()]))
    )
    proc = await asyncio.create_subprocess_exec(
        *args, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
    )
    stdout, stderr = await proc.communicate()

    # If proc exits with a non-zero exit code, the stdout may still
    # be valid JSON, but we expect it to have an "errors" property defined.
    try:
        response = json.loads(stdout)
    except json.JSONDecodeError:
        response = None

    if proc.returncode == 0:
        assert response is not None
        assert "errors" not in response
        return Result(ok=response)
    elif response is not None:
        return Result(error=json.dumps(response, indent=1))
    elif b"gh auth login" in stderr:
        # The error message is likely referring to an authentication issue.
        raise error.Abort(_("Error calling the GitHub API:\n%s") % stderr.decode())
    else:
        return Result(
            error=f"exit({proc.returncode}) Failure running {' '.join(args)}\nstdout: {stdout.decode()}\nstderr: {stderr.decode()}\n"
        )


def _format_param(key: str, value: Union[str, int, bool]) -> List[str]:
    # In Python, bool is a subclass of int, so check it first.
    if isinstance(value, bool):
        opt = "-F"
        val = str(value).lower()
    elif isinstance(value, int):
        opt = "-F"
        val = value
    elif isinstance(value, str):
        opt = "-f"
        val = str(value)
    else:
        raise RuntimeError(f"unexpected param: {key}={value}")
    return [opt, f"{key}={val}"]
